/* библиотеки, используемые для подавления BrownOut-сброса */
#include "soc/soc.h"            // устанавливается автоматом как зависимость к библиотеке STAMP-PICO
#include "soc/rtc_cntl_reg.h"   // устанавливается автоматом как зависимость к библиотеке STAMP-PICO

#include <WiFi.h>               // устанавливается автоматом как зависимость к библиотеке STAMP-PICO
#include <AsyncTCP.h>           // установка с сайта github -> https://github.com/me-no-dev/AsyncTCP 
#include <ESPAsyncWebServer.h>  // установка с сайта github -> https://github.com/me-no-dev/ESPAsyncWebServer 
//#include <AsyncElegantOTA.h>  // установка через менеджер библиотек: http://librarymanager/All#AsyncElegantOTA
#include <ElegantOTA.h>         // установка с сайта github -> https://github.com/ayushsharma82/ElegantOTA

#include <LittleFS.h>           // устанавливается автоматом как зависимость к библиотеке STAMP-PICO
#include <ESPDateTime.h>        // установка через менеджер библиотек: http://librarymanager/All#ESPDateTime
#include <FastLED.h>            // устанавливается автоматом как зависимость к библиотеке STAMP-PICO
#include <Preferences.h>        // установка через менеджер библиотек: http://librarymanager/All#Preferences
#include <PubSubClient.h>       // установка через менеджер библиотек: http://librarymanager/All#PubSubClient  (by Nick O'Leary)
#include <CircularBuffer.h>     // установка через менеджер библиотек: http://librarymanager/All#CircularBuffer (by Roberto Lo Giacco)
#include <OneWire.h>            // установка через менеджер библиотек: http://librarymanager/All#OneWire (by Paul Stoffregen)
#include <DallasTemperature.h>  // установка через менеджер библиотек: http://librarymanager/All#DallasTemperature (by Miles Burton)

AsyncWebServer server(80);
/* https://randomnerdtutorials.com/esp32-save-data-permanently-preferences/ */
Preferences preferences;
//===============================================================
/* просто идентификатор чипа, используестя при отправке логов */
String mac2String(byte ar[]) {
  String s;
  for (byte i = 0; i < 6; ++i)
  {
    char buf[3];
    sprintf(buf, "%02X", ar[i]); // J-M-L: slight modification, added the 0 in the format for padding
    s += buf;
    //  if (i < 5) s += '-';
  }
  return s;
}
uint64_t espChipID = ESP.getEfuseMac();
String chipId = "HTR-" + String(mac2String((byte*) &espChipID)) ;
//===============================================================
/*
   Плата использует UDP-протокол для отправки сообщений на внешний SYSLOG-
   сервер. Номер порта - фиксированный, стандартный для SYSLOG - 514-й.
   Адрес сервера - указывается пользователем в настройках через веб-интерфейс
   и затем хранится в энзу.
*/
#include <WiFiUdp.h>          // устанавливается автоматом как зависимость к библиотеке STAMP-PICO
String udpAddress = "";
const int udpPort = 514;
WiFiUDP udp;
String logMessageToSend;
//===============================================================
/*
   "структура" с параметрами, используемыми расписанием работы:
   sched_rows_numb  - максимальное число записей (размер массива)
   single_sched_record  - структура одной записи
   work_shedule - весь массив записей (инициализация)
*/
const uint16_t sched_rows_numb = 30;
typedef struct
{
  uint16_t  start_time;
  uint16_t  stop_time;
  float     desired_temp;
  bool      isactiv;
}  single_sched_record;
single_sched_record work_shedule[sched_rows_numb];
//===============================================================
/*
    Массивы с текстовыми сообщениями
*/
const PROGMEM char daysOfTheWeek[7][5] = {"ВС", "ПН", "ВТ", "СР", "ЧТ", "ПТ", "СБ"};
const PROGMEM char availableModes[3][21] = {"НАГРЕВ", "ОХЛАЖДЕНИЕ", "РАСПИСАНИЕ"};
const PROGMEM char powerStates[2][24] = {"ВЫКЛЮЧЕНО", "ВКЛЮЧЕНО"};
const PROGMEM char errorMessagesArray [10][181] {
  {"АВАРИЯ - Нет сигнала от основного датчика температуры! Оборудование выключено!"}, // 0
  {"АВАРИЯ - Нет сигнала от аварийного датчика температуры! Оборудование выключено!"},  // 1
  {"АВАРИЯ - Температура превысила аврийный порог! Оборудование выключено!"}, // 2
  {"Загрузка системы завершена! Был получен такой IP адрес: "}, // 3
  {"Поступила команда принудительного включения оборудования! Оно будет включено в течение "},  // 4
  {"Оборудование сейчас невозможно включить принудительно, т.к. система находится в состоянии ошибки!"}, // 5
  {"Согласно расписания ВКЛЮЧЕН режим поддержания заданной температуры: "},  // 6
  {"Согласно расписания режим поддержания заданной температуры ВЫКЛЮЧЕН. Текущ. темп-ра: "},  // 7
  {"Активное оборудование ВЫКЛЮЧЕНО, т.к. температура достигла требуемого значения: "},  // 8
  {"Активное оборудование ВКЛЮЧЕНО, т.к. температура вышла за допустимые пределы: "}  //9
};
//===============================================================
/*
   Implementing ESP32 hardware watchdog timer:
   first string - library, and second - set delay
   in seconds after which ESP32 will reset
*/
#include <esp_task_wdt.h>          // устанавливается автоматом как зависимость к библиотеке STAMP-PICO
#define WDT_TIMEOUT 300
//===============================================================
/* регистр флагов ошибок */
uint8_t errors_flags = 0;
/* флаги ошибок (биты с соотв. номерами, хранящиеся в регистре ошибок) */
#define NO_WIFI           0
#define NO_NTP            1
#define LTFS_FAIL         2
#define NO_TEMP_SENS_W    3
#define NO_TEMP_SENS_A    4
#define BOILER_OVERHEAT   7
//===============================================================
/* настройки для датчиков температуры DS18B20 */
#define temperature_pin_1   18
OneWire sensor_1(temperature_pin_1);
DallasTemperature main_sensor(&sensor_1); // "рабочий"
#define temperature_pin_2   26
OneWire sensor_2(temperature_pin_2);
DallasTemperature alarm_sensor(&sensor_2); // "аварийный"
//===============================================================
/* парметры для связи с сервером mqtt */
String mqttserver;
String mqttuser;
String mqttpassword;
#define MQTT_PUB_LOG        "logrecords"
#define MQTT_PUB_TEMP       "temperature"
#define MQTT_PUB_POWERSTATE "powerstate"
#define MQTT_PUB_FORCE_ON   "forceon"
const int mqtt_port = 1883;
String mqttMessageToSend;
bool mqttCredentialsPresent = false;
/* create an instance of PubSubClient client */
WiFiClient espClient;
PubSubClient mqtt_client;
//===============================================================
/* this below - used for sending data to site-collector */
String logger;
String collectorSiteAddress = "/termo/parse.php?";
#include <HTTPClient.h>          // устанавливается автоматом как зависимость к библиотеке STAMP-PICO
HTTPClient http;
//===============================================================
/* счетчики временных интервалов (1 минута и 10 минут) */
uint8_t one_min_counter = 0;
uint8_t ten_min_counter = 0;
//===============================================================
/* настройки для FIFO-буфера (используется для сбора и отправки лог-сообщений */
CircularBuffer < byte, 100> messages_quene;
//===============================================================
/* логин и пароль для "обновления прошивки по воздуху" */
const char* ota_username = "firmware";
const char* ota_password = "firmware";
//===============================================================
/* параметры системы */
/* "Имя" устройства */
String devicename;
/* Счетчики времени общего работы исполнительного элемена (например, нагревателя) */
uint64_t powerOnTimeTotal = 0;
uint64_t powerOnTimePrevious = 0;
/* текущие показания датчиков температуры */
float currentTempWorking = 0.0;
float currentTempAlarm = 0.0;
/* параметр, указывающий текущий режим работы системы (нагрев(0)/охлаждение(1)/расписание(2)) */
uint8_t mainMode = 0;
/* гистерезис поддержания температуры */
float hysteresis = 1.0;
//===============================================================
/* флаги, указывающие на необходимости записи параметров в ЕЕПРОМ */
bool needSaveTotalTime = false;
bool needSaveCredentials = false;
bool needSaveMqttCredentials = false;
bool needSaveSchedule = false;
bool needSaveSettings = false;
//===============================================================
/* флаг того, что система только что загрузилась */
bool bootAfterPowerLost = true;
//===============================================================
/* флаг прерывания, срабатывающего каждые 10 секунд */
volatile bool tenSecondFlag = false;
/* номер задания из шедуллера, которое выполняется в данный момент (если "-1", то никакое)*/
int8_t selectedScheduleRow = -1;
/* флаг, разрешающий работу исполнительного механизма */
bool powerControlAllowed = false;
/* флаг показывающий реальное текущее состояние исполнительного механизма */
bool powerState = false;
/* счетчик таймера, по которому активируется принудительное включение исполнительного механизма */
uint16_t forceOnTimerCounter = 0;
/* шаг (в секундах) на который увеличивается счетчик времени принудительного включения  */
#define FORCE_ON_TIMER_STEP 7200
/* температура, поддерживаемая во время режима принудительного включения */
float forcetemp;
/* температура аварийного перегрева */
float alarmtemp;
/* октеты статического IP-адреса, который система будет пытаться установить себе  */
uint16_t ip0 , ip1 , ip2 , ip3 ;
/* текущее сообщение об ошибке отправляемое в браузер */
String errorMessageText;
/* параметры собственной беспроводной точки доступа (если не подключен к внешним АР) */
String ap_ssid = "Stamp_Pico";
String ap_password = "12345678";
/* параметры для подключения к беспроводной сети с внешней АР (будут считаны из энзу) */
String ssid ;
String password ;
/* настройки библиотеки FastLED для управления многоцветным светодиодом */
#define NUM_LEDS 1
#define DATA_PIN 27
CRGB leds[NUM_LEDS];
/* вывод, используемый для управления реле, включающим исполнительный механизм */
#define relay_control_pin   25
/* вывод, используемый для кнопки включения принудительного режима */
#define force_button_pin   19
//===============================================================
//          П Р Е Р Ы В А Н И Я
//===============================================================
/*
   Прерывание 0-го таймера. Настроено так, чтобы происходить раз в 10 секунд
   По прерыванию выставляется флаг (oneSecondFlag) для основной программы
*/
hw_timer_t * timer0 = NULL;
portMUX_TYPE timer0Mux = portMUX_INITIALIZER_UNLOCKED;
void IRAM_ATTR onTimer0() {
  portENTER_CRITICAL_ISR(&timer0Mux);
  tenSecondFlag = true;
  portEXIT_CRITICAL_ISR(&timer0Mux);
}

void setup() {
  /* подавление BrownOut-сброса */
  WRITE_PERI_REG(RTC_CNTL_BROWN_OUT_REG, 0);

  /* настройки входов/выходов */
  pinMode(temperature_pin_1, INPUT_PULLUP);
  pinMode(temperature_pin_2, INPUT_PULLUP);
  digitalWrite(relay_control_pin, LOW);
  pinMode(relay_control_pin, OUTPUT);
  pinMode(force_button_pin, INPUT_PULLUP);

  /* инициализируем монитор COM-порта */
  Serial.begin(115200);
  while (!Serial) {
    ; /* ожидаем подключения последовательного порта */
  }
  /* управление разноцветным светодиодом, установленным на самой плате */
  /* инициализация самой библиотеки управления светодиодом (GRB ordering is typical)*/
  FastLED.addLeds<SK6812, DATA_PIN, RGB>(leds, NUM_LEDS);
  /* включаем красный цвет (на время загрузки и инициализации проца) */
  setLedColorRed();

  /* инициализируем файловую систему LittleFS (в ней хранятся файлы веб-сервера) */
  if (LittleFS.begin(true)) {
    server.serveStatic("/files/", LittleFS, "/files/", "max-age=86400");
  } else {
    Serial.println(F("An Error has occurred while mounting LittleFS"));
    bitWrite(errors_flags, LTFS_FAIL, HIGH);
  }

  /* считываем из preferences (хранилища) основные настройки системы) */
  /* инициализация */
  preferences.begin("all-settings", true); // true - только чтение, false - и чтение и запись
  /*Считываем из preferences SSID беспроводной сети */
  ssid = preferences.getString("ssid", "");
  /* и выводим его в консоль */
  Serial.printf("WiFi SSID from Preferences is: \"%s\".\n", ssid.c_str());
  /* считываем из preferences пароль беспроводной сети */
  password = preferences.getString("password", "");
  /* и выводим его в консоль */
  Serial.printf("WiFi Password from Preferences is: \"%s\".\n", password.c_str());
  /* считываем из preferences октеты IP-адреса беспроводной сети */
  ip0 = preferences.getUShort("ip0", 0);
  ip1 = preferences.getUShort("ip1", 0);
  ip2 = preferences.getUShort("ip2", 0);
  ip3 = preferences.getUShort("ip3", 0);
  Serial.printf("Desired static IP-address from Preferences is: %d.%d.%d.%d \n", ip0, ip1, ip2, ip3);
  /* считываем из preferences адрес SYSLOG-сервера */
  udpAddress = preferences.getString("udpAddress", "");
  Serial.printf("External Syslog server address is: \"%s\".\n", udpAddress.c_str());
  /* считываем из preferences имя устройства */
  devicename = preferences.getString("devicename", "");
  /* считываем из preferences адрес сервера логгирования температуры */
  logger = preferences.getString("logger", "");
  /* считываем из preferences параметры для связи с mqtt-сервером */
  mqttserver = preferences.getString("mqttserver", "");
  mqttuser = preferences.getString("mqttuser", "");
  mqttpassword = preferences.getString("mqttpassword", "");
  Serial.printf("Параметры для связи с mqtt-сервером: сервер - %s, юзер - %s, пароль - %s.\n", mqttserver.c_str(), mqttuser.c_str(), mqttpassword.c_str());
  if (!mqttserver.isEmpty() && !mqttuser.isEmpty() && !mqttpassword.isEmpty()) {
    mqttCredentialsPresent = true;
  }
  /* настраиваем MQTT-клиент  */
  if (mqttCredentialsPresent) {
    mqtt_client.setClient(espClient);
    mqtt_client.setServer(mqttserver.c_str(), mqtt_port);
    mqtt_client.setCallback(receivedCallback);
    mqtt_client.setBufferSize(512);
  }
  /* считываем из preferences основной режим работы (нагрев или охраждение) */
  mainMode = preferences.getUChar("mainMode", 0);
  Serial.printf("Режим работы системы: \"%s\".\n", availableModes[mainMode]);
  /* считываем из preferences гистерезис (порого) поддержания темпертуры */
  hysteresis = preferences.getFloat("hysteresis", 1.00);
  Serial.print(F("Гистерезис поддержания температуры: "));
  Serial.println(hysteresis);
  forcetemp = preferences.getFloat("forcetemp", 20.00);
  alarmtemp = preferences.getFloat("alarmtemp", 80.00);
  powerOnTimeTotal = preferences.getULong64("powercntr", 0);
  if (powerOnTimeTotal != 0) {
    powerOnTimePrevious = powerOnTimeTotal;
  }
  Serial.print(F("total work time from eeprom: "));
  Serial.println(formattedTimeFromInt(powerOnTimeTotal));
  /* считываем из preferences (хранилища) "расписание работы" */
  read_shedule_from_prefs();
  preferences.end();

  wifiConnectionSetup();

  /* -------------------------- настройки веб-сервера --------------------------- */
  /* URL для корневой страницы веб-сервера  */
  server.on("/", HTTP_GET, [](AsyncWebServerRequest * request) {
    request->send(LittleFS, "/index.html");
  });

  /* отвен на POST-запрос, поступающий по адресу "/setwifi"
     (=> запись в Preferences параметров для подключения к WiFi) */
  server.on("/setwifi", HTTP_POST, [](AsyncWebServerRequest * request) {

    bool vrongSomeValue = false;
    String value1, value2, value3, value4, value5, value6, value7, value8;

    if (request->hasParam("ssid", true)) {
      value2 = request->arg("ssid");
    }
    if (request->hasParam("password", true)) {
      value3 = request->arg("password");
    }
    if (request->hasParam("ip3", true)) {
      value4 = request->arg("ip3");
    }
    if (request->hasParam("ip2", true)) {
      value5 = request->arg("ip2");
    }
    if (request->hasParam("ip1", true)) {
      value6 = request->arg("ip1");
    }
    if (request->hasParam("ip0", true)) {
      value7 = request->arg("ip0");
    }
    if (request->hasParam("syslog", true)) {
      value8 = request->arg("syslog");
    }

    if (value2.length() > 32) {
      errorMessageText = F("Длинна имени беспроводной сети превышает 32 символа!");
      vrongSomeValue = true;
    } else if (password.length() > 32) {
      errorMessageText = F("Длинна пароля превышает 32 символа!");
      vrongSomeValue = true;
    } else if (value4.toInt() < 0 || value4.toInt() > 255) {
      errorMessageText = F("Указано неверное значение для IP-адреса!");
      vrongSomeValue = true;
    } else if (value5.toInt() < 0 || value5.toInt() > 255) {
      errorMessageText = F("Указано неверное значение для IP-адреса!");
      vrongSomeValue = true;
    } else if (value6.toInt() < 0 || value6.toInt() > 255) {
      errorMessageText = F("Указано неверное значение для IP-адреса!");
      vrongSomeValue = true;
    } else if (value7.toInt() < 0 || value7.toInt() > 255) {
      errorMessageText = F("Указано неверное значение для IP-адреса!");
      vrongSomeValue = true;
    } else {
      errorMessageText = "Сейчас будет выполнен перезапуск системы, ждите...";
    }

    if (!vrongSomeValue) {
      ssid = value2;
      password = value3;
      ip3 = value4.toInt();
      ip2 = value5.toInt();
      ip1 = value6.toInt();
      ip0 = value7.toInt();
      if (!value8.isEmpty()) {
        udpAddress = value8;
      }
      needSaveCredentials = true;
    }
    /* и теперь переадресуем клиента на "корневую" страницу сервера */
    request->redirect("/");
  });

  server.on("/setmqtt", HTTP_POST, [](AsyncWebServerRequest * request) {
    bool vrongSomeValue = false;
    String value1, value2, value3, value4;
    if (request->hasParam("mqttserver", true)) {
      value1 = request->arg("mqttserver");
    }
    if (request->hasParam("mqttuser", true)) {
      value2 = request->arg("mqttuser");
    }
    if (request->hasParam("mqttpassword", true)) {
      value3 = request->arg("mqttpassword");
    }
    if (request->hasParam("logger", true)) {
      value4 = request->arg("logger");
    }

    if (value1.length() > 32) {
      errorMessageText = F("Длинна адреса MQTT-сервера превышает 32 символа!");
      vrongSomeValue = true;
    } else if (value2.length() > 32) {
      errorMessageText = F("Длинна имени пользователя для связи с MQTT-сервером превышает 32 символа!");
      vrongSomeValue = true;
    } else if (value3.length() > 32) {
      errorMessageText = F("Длинна пароля для связи с MQTT-сервером превышает 32 символа!");
      vrongSomeValue = true;
    } else if (value4.length() > 64) {
      errorMessageText = F("Длинна адреса сервера сбора данных превышает 64 символа!");
      vrongSomeValue = true;
    }

    if (!vrongSomeValue) {
      mqttserver = value1;
      mqttuser = value2;
      mqttpassword = value3;
      logger = value4;
      needSaveMqttCredentials = true;
    }
    /* и теперь переадресуем клиента на "корневую" страницу сервера */
    request->redirect("/");
  });

  server.on("/setmain", HTTP_POST, [](AsyncWebServerRequest * request) {
    bool vrongSomeValue = false;
    String value1, value2, value3, value4, value5;

    if (request->hasParam("main_mode", true)) {
      value1 = request->arg("main_mode");
    }
    if (request->hasParam("hysteresis", true)) {
      value2 = request->arg("hysteresis");
    }
    if (request->hasParam("forcetemp", true)) {
      value3 = request->arg("forcetemp");
    }
    if (request->hasParam("alarmtemp", true)) {
      value4 = request->arg("alarmtemp");
    }
    if (request->hasParam("devname", true)) {
      value5 = request->arg("devname");
    }

    if (value1.isEmpty()) {
      errorMessageText = F("Вы не указали желаемый режим работы!");
      vrongSomeValue = true;
    } else if (value2.isEmpty()) {
      errorMessageText = F("Вы не указали значение гистерезиса!");
      vrongSomeValue = true;
    } else if (!isNumeric(value2)) {
      errorMessageText = F("Значение гистерезиса должно быть числом!");
      vrongSomeValue = true;
    } else if (value3.isEmpty()) {
      errorMessageText = F("Вы не указали значение температуры принудительного режима!");
      vrongSomeValue = true;
    } else if (!isNumeric(value3)) {
      errorMessageText = F("Значение температуры принудительного режима должно быть числом!");
      vrongSomeValue = true;
    } else if (value4.isEmpty()) {
      errorMessageText = F("Вы не указали значение температуры аварийного предела!");
      vrongSomeValue = true;
    } else if (!isNumeric(value4)) {
      errorMessageText = F("Значение температуры аварийного предела должно быть числом!");
      vrongSomeValue = true;
    } else if (value1.toInt() < 0 || value1.toInt() > 2) {
      errorMessageText = F("Вы указали неверный режим работы!");
      Serial.println(value1.toInt());
      vrongSomeValue = true;
    } else if (value2.toFloat() < 0 || value1.toFloat() > 100) {
      errorMessageText = F("Вы указали неприемлемое значение гистерезиса!");
      vrongSomeValue = true;
    }
    if (value3.toFloat() < -100.00 || value3.toFloat() > 100.00 ) {
      vrongSomeValue = true;
      errorMessageText = F("Указаное значение температуры выходит за допустимые пределы!");
    }
    if (value4.toFloat() < -100.00 || value4.toFloat() > 100.00 ) {
      vrongSomeValue = true;
      errorMessageText = F("Указаное значение температуры аварийного перегрева выходит за допустимые пределы!");
    }
    if (!vrongSomeValue) {
      mainMode = value1.toInt();
      hysteresis = value2.toFloat();
      forcetemp = value3.toFloat();
      alarmtemp = value4.toFloat();
      devicename = value5;
      needSaveSettings = true;
    }
    /* и теперь переадресуем клиента на "корневую" страницу сервера */
    request->redirect("/");
  });

  server.on("/setschedule", HTTP_POST, [](AsyncWebServerRequest * request) {
    bool vrongSomeValue = false;
    bool isactiv = false;
    uint8_t row = 200;
    uint8_t start_day, start_hour, start_minute, stop_day, stop_hour, stop_minute;
    float temperature = -127.0 ;
    if (request->hasParam("row", true)) {
      row = request->arg("row").toInt();
    }
    if (row >= sched_rows_numb) {
      errorMessageText = F("Запрос не содержит необходимой информации!");
    } else {
      if (request->hasParam("start_day", true)) {
        start_day = request->arg("start_day").toInt();
      }
      if (request->hasParam("start_hour", true)) {
        start_hour = request->arg("start_hour").toInt();
      }
      if (request->hasParam("start_minute", true)) {
        start_minute = request->arg("start_minute").toInt();
      }
      if (request->hasParam("stop_day", true)) {
        stop_day = request->arg("stop_day").toInt();
      }
      if (request->hasParam("stop_hour", true)) {
        stop_hour = request->arg("stop_hour").toInt();
      }
      if (request->hasParam("stop_minute", true)) {
        stop_minute = request->arg("stop_minute").toInt();
      }
      if (request->hasParam("temperature", true)) {
        temperature = request->arg("temperature").toFloat();
      }
      if (request->hasParam("isactiv", true)) {
        isactiv = true;
      }
    }

    if (start_day < 0 || start_day > 6 ) {
      vrongSomeValue = true;
      errorMessageText = F("Значение дня недели начала интервала выходит за допустимые пределы!");
    }
    if (start_hour < 0 || start_hour > 23 ) {
      vrongSomeValue = true;
      errorMessageText = F("Значение часов начала интервала выходит за допустимые пределы!");
    }
    if (start_minute < 0 || start_minute > 59 ) {
      vrongSomeValue = true;
      errorMessageText = F("Значение минут начала интервала выходит за допустимые пределы!");
    }
    if (stop_day < 0 || stop_day > 6 ) {
      vrongSomeValue = true;
      errorMessageText = F("Значение дня недели окончания интервала выходит за допустимые пределы!");
    }
    if (stop_hour < 0 || stop_hour > 23 ) {
      vrongSomeValue = true;
      errorMessageText = F("Значение часов окончания интервала выходит за допустимые пределы!");
    }
    if (stop_minute < 0 || stop_minute > 59 ) {
      vrongSomeValue = true;
      errorMessageText = F("Значение часов окончания интервала выходит за допустимые пределы!");
    }
    if (temperature < -20.00 || temperature > 99.00 ) {
      vrongSomeValue = true;
      errorMessageText = F("Указаное значение температуры выходит за допустимые пределы!");
    }

    if (!vrongSomeValue) {
      /* записываем новые значения в массив расписания (в ОЗУ) */
      work_shedule[row].start_time = anyDayTimeToInt(start_day, start_hour, start_minute);
      work_shedule[row].stop_time = anyDayTimeToInt(stop_day, stop_hour, stop_minute);
      work_shedule[row].desired_temp = temperature;
      work_shedule[row].isactiv = isactiv;
      /* формируем текстовое сообщение в браузер с новыми значениями */
      errorMessageText = "Новые значения для задания No ";
      errorMessageText += (row + 1);
      errorMessageText += ": начало интервала в ";
      errorMessageText += intCurDayTimeToString(anyDayTimeToInt(start_day, start_hour, start_minute));
      errorMessageText += ", окончание &mdash; в ";
      errorMessageText += intCurDayTimeToString(anyDayTimeToInt(stop_day, stop_hour, stop_minute));
      errorMessageText += ", поддерживаемая температура ";
      errorMessageText += temperature;
      errorMessageText += "&#8451;";
      /* ставим флаг "нужно сохранить расписание" */
      needSaveSchedule = true;
    }
    request->redirect("/");
  });

  server.on("/getscedule", HTTP_GET, [](AsyncWebServerRequest * request) {
    String schedule_table = "<table>\n<tr>\n<th>No</th>\n<th>Старт</th>\n<th>Стоп</th>\n<th>Температура</th>\n<th>Редакт.</th>\n</tr>\n";
    for (uint8_t i = 0 ; i < sched_rows_numb ; i++ ) {
      String row_style = ( work_shedule[i].isactiv != true || work_shedule[i].start_time == work_shedule[i].stop_time ) ? " style=\"color:#b1a8a8;\"" : "";
      schedule_table += "<tr" + row_style + " id=\"taskrow" + (1 + i) + "\">\n<td>";
      schedule_table += (i + 1);
      schedule_table += "</td>\n<td>";
      schedule_table += intCurDayTimeToString(work_shedule[i].start_time) ;
      schedule_table += "</td>\n<td>";
      schedule_table += intCurDayTimeToString(work_shedule[i].stop_time);
      schedule_table += "</td>\n<td>";
      schedule_table += work_shedule[i].desired_temp;
      schedule_table += "&#8451;</td>\n<td title=\"Изменить\">";
      schedule_table += "<div class=\"pencil\" onClick=\"editrow(";
      schedule_table += i;
      schedule_table += ");\">&#10000;</div>";
      schedule_table += "</td>\n</tr>\n";
    }
    schedule_table += "</table>";
    request->send(200, "text/html", schedule_table);
  });

  server.on("/fillform", HTTP_GET, [](AsyncWebServerRequest * request) {
    String whatToSend;
    bool onlyPopUp = false;
    String json = "{";
    if (request->hasParam("formname")) {
      whatToSend = request->getParam("formname")->value();
    }
    if (whatToSend.isEmpty()) {
      json += "\"error\":\"Empty request!\"";

    } else if (whatToSend == "settingsbox") {
      json += "\"form_main_mode\":\"";
      json += mainMode;
      json += "\",\"hysteresis\":\"";
      json += hysteresis;
      json += "\",\"alarmtemp\":\"";
      json += alarmtemp;
      json += "\",\"forcetemp\":\"";
      json += forcetemp;
      json += "\",\"devname\":\"";
      json += devicename;
      json += "\"";

    } else if (whatToSend == "power_state") {
      onlyPopUp = true;
      errorMessageText = "Общее время активной работы оборудования составляет ";
      errorMessageText += formattedTimeFromInt(powerOnTimeTotal);

    } else if (whatToSend == "showforce") {
      onlyPopUp = true;
      errorMessageText = "До окончания принудительного режима осталось ";
      errorMessageText += formattedTimeFromInt(forceOnTimerCounter);

    } else if (whatToSend == "wifi_scan") {
      onlyPopUp = true;
      uint8_t nn = WiFi.scanNetworks();
      if (nn == 0) {
        errorMessageText = "Беспроводные сети не найдены!";
      } else {
        errorMessageText = "***** Найдено ";
        errorMessageText += nn;
        errorMessageText += " беспроводных сетей *****<br>";
        for (int i = 0; i < nn; ++i) {
          errorMessageText += (i + 1);
          errorMessageText += ": &quot;";
          errorMessageText += WiFi.SSID(i);
          errorMessageText += "&quot;, уровень сигнала: ";
          errorMessageText += WiFi.RSSI(i);
          errorMessageText += " dBi<br>";
        }
      }

    } else if (whatToSend == "mqttbox") {
      json += "\"mqttserver\":\"";
      json += mqttserver;
      json += "\",\"mqttuser\":\"";
      json += mqttuser;
      json += "\",\"mqttpassword\":\"";
      json += mqttpassword;
      json += "\",\"logger\":\"";
      json += logger;
      json += "\"";

    } else if (whatToSend == "wifibox") {
      json += "\"ssid\":\"";
      json += ssid;
      json += "\",\"password\":\"";
      json += password;
      json += "\",\"ip0\":\"";
      json += ip0;
      json += "\",\"ip1\":\"";
      json += ip1;
      json += "\",\"ip2\":\"";
      json += ip2;
      json += "\",\"ip3\":\"";
      json += ip3;
      json += "\",\"syslog\":\"";
      json += udpAddress;
      json += "\"";

    } else if (whatToSend == "mainpage") {
      json += "\"curtime\":\"";
      json += DateTime.toString().c_str() ;
      json += "\",\"devicename\":\"";
      if (!devicename.isEmpty()) {
        json += " (" + devicename + ")";
      }
      json += "\",\"main_mode\":\"";
      if (errors_flags > 0) {
        json += "<span class='alarm' onClick='showhelp();'>АВАРИЯ : ";
        for (int i = 7; i >= 0; i--) {
          if (bitRead(errors_flags, i) == 1) {
            json += "1";
          } else {
            json += "0";
          }
        }
        json += "</div>";
      } else {
        if (forceOnTimerCounter > 0) {
          json += "<span class='working'><div id='showforce'>ПРИНУД. </div>";
        }
        json += availableModes[mainMode] ;
      }
      json += "\",\"curr_temp\":\"";
      if (mainMode != 2) {
        json += currentTempWorking ;
        json += " / ";
        json += currentTempAlarm ;
      } else {
        json += "..." ;
      }

      json += "\",\"power_state\":\"";
      if (powerState == 1) {
        json += "<span class='working'>";
      }
      json += powerStates[powerState] ;
      json += "\",\"row_no\":\"";
      if (selectedScheduleRow < 0 ) {
        json += "&mdash;" ;
      } else {
        json += (selectedScheduleRow + 1);
      }
      json += "\"";
    }

    if (errorMessageText.length() > 0) {
      if (!onlyPopUp) {
        json += ",";
      }
      json += "\"error\":\"";
      json += errorMessageText;
      json += "\"";
      /* обнуляем само сообщение */
      errorMessageText = "";
    }
    json += "}";
    request->send(200, "application/json", json);
  });

  server.on("/totalcounerreset", HTTP_GET, [](AsyncWebServerRequest * request) {
    powerOnTimeTotal = 0;
    needSaveTotalTime = true;
    request->redirect("/");
  });

  server.on("/schedrow", HTTP_GET, [](AsyncWebServerRequest * request) {
    String json = "{";
    uint8_t rownumber = 200;
    if (request->hasParam("rownumber")) {
      rownumber = request->getParam("rownumber")->value().toInt();
    }
    if (rownumber == 200) {
      json += "\"error\":\"Empty request!\"";
    } else {
      uint16_t start_time = work_shedule[rownumber].start_time;
      uint16_t stop_time = work_shedule[rownumber].stop_time;

      uint8_t day_start = start_time / 24 / 60 ;
      uint8_t hour_start = (start_time / 60) % 24;
      uint8_t minute_start = start_time % 60;

      uint8_t day_stop = stop_time / 24 / 60 ;
      uint8_t hour_stop = (stop_time / 60) % 24;
      uint8_t minute_stop = stop_time % 60;

      json += "\"start_day\":\"";
      json += day_start ;
      json += "\",\"start_hour\":\"";
      if (hour_start < 10) {
        json += "0" ;
      }
      json += hour_start ;
      json += "\",\"start_minute\":\"";
      if (minute_start < 10) {
        json += "0" ;
      }
      json += minute_start ;
      json += "\",\"stop_day\":\"";
      json += day_stop ;
      json += "\",\"stop_hour\":\"";
      if (hour_stop < 10) {
        json += "0" ;
      }
      json += hour_stop ;
      json += "\",\"stop_minute\":\"";
      if (minute_stop < 10) {
        json += "0" ;
      }
      json += minute_stop ;
      json += "\",\"temperature\":\"";
      json += work_shedule[rownumber].desired_temp ;
      json += "\",\"isactiv\":\"";
      json += work_shedule[rownumber].isactiv ;
      json += "\"";
    }
    json += "}";
    request->send(200, "application/json", json);
  });

  server.onNotFound([](AsyncWebServerRequest * request) {
    request->send(404);
  });

  /* -------------------------- конец настроек веб-сервера --------------------------- */

  /* Запуск ElegantOTA ( => возможность обновления прошивки "по воздуху") */
  //AsyncElegantOTA.begin(&server, ota_username, ota_password);
  ElegantOTA.setAuth(ota_username, ota_password);
  ElegantOTA.begin(&server);


  /* запускаем веб-сервер */
  server.begin();

  /* Запуск таймера, который будует давать прерывание раз в 10 секунд */
  /* https://techtutorialsx.com/2017/10/07/esp32-arduino-timer-interrupts/ */
  timer0 = timerBegin(0, 80, true); // тут: 0 - номер таймера, 80 - прескаллер, true - постоянная работа
  timerAttachInterrupt(timer0, &onTimer0, true); // подключаем прерывание
  timerAlarmWrite(timer0, 10000000, true);  // ставим делитель такой, чтобы прерывание происходило каждые 10 с
  timerAlarmEnable(timer0); // включаем сам таймер

  main_sensor.begin();
  alarm_sensor.begin();

  /* this two strings below used for setup and enable watchdog timer of ESP32 */
  esp_task_wdt_init(WDT_TIMEOUT, true); //enable panic so ESP32 restarts
  esp_task_wdt_add(NULL); //add current thread to WDT watch

}

boolean isNumeric(String str) {
  unsigned int stringLength = str.length();
  if (stringLength == 0) {
    return false;
  }
  boolean seenDecimal = false;
  for (unsigned int i = 0; i < stringLength; ++i) {
    if (isDigit(str.charAt(i))) {
      continue;
    }
    if (str.charAt(i) == '.') {
      if (seenDecimal) {
        return false;
      }
      seenDecimal = true;
      continue;
    }
    return false;
  }
  return true;
}

void read_shedule_from_prefs() {
  size_t totSchLen = preferences.getBytesLength("work-schedule");
  char buffer[totSchLen]; // prepare a buffer for the data
  preferences.getBytes("work-schedule", buffer, totSchLen);
  if (totSchLen % sizeof(work_shedule)) { // simple check that data fits
    Serial.println(F("Data is not correct size!"));
    return;
  }
  memcpy(work_shedule, buffer, totSchLen);
  Serial.println(F("Reading schedule records from preferences..."));
  Serial.printf("Size of schedule at EEPROM is %d and size of schedule array is %d\n", totSchLen, sizeof(work_shedule));
}

/* основной цикл */
void loop() {

  /* отправка сообщений на сервер SYSLOG */
  if (!logMessageToSend.isEmpty() && !udpAddress.isEmpty() && WiFi.status() == WL_CONNECTED && !bootAfterPowerLost) {
    syslog_message_send();
  }

  /* отправка сообщений на сервер MQTT */
  if (!logMessageToSend.isEmpty() && mqttCredentialsPresent && WiFi.status() == WL_CONNECTED && !bootAfterPowerLost) {
    uint8_t result = mqtt_log_message_send();
    if (result > 0) {
      mqttMessageToSend = String();
    }
  }

  /* проверка очереди лог-сообщений и подготовка их к отправке на сервера MQTT и SYSLOG */
  if (!messages_quene.isEmpty()) {
    uint8_t messageIndex = messages_quene.first();
    mqttMessageToSend = errorMessagesArray[messageIndex];
    if (messageIndex == 3) {
      mqttMessageToSend += WiFi.localIP().toString();
    } else if (messageIndex == 4) {
      mqttMessageToSend += formattedTimeFromInt(forceOnTimerCounter);
    } else if (messageIndex == 6) {
      mqttMessageToSend += work_shedule[selectedScheduleRow].desired_temp;
      mqttMessageToSend += "ºC";
    } else if (messageIndex >= 7) {
      mqttMessageToSend += currentTempWorking;
      mqttMessageToSend += "ºC";
    }
    logMessageToSend = mqttMessageToSend;
    Serial.println(mqttMessageToSend);
    messages_quene.shift();
  }

  /* вызов по флагу процедуры выполняемой каждые 10 секунд */
  if (tenSecondFlag) {
    tenSecondProcedure();
  }

  /* вызов по флагам процедур записи в энзу (flash) измененных параметров */
  if (needSaveCredentials) {
    saveWifiCredentials();
  } else if (needSaveSchedule) {
    saveScheduleRecords();
  } else if (needSaveSettings) {
    saveMainSettings();
  } else if (needSaveMqttCredentials) {
    saveMqttCredentials();
  } else if (needSaveTotalTime) {
    savePowerConsumption();
  }

  /* "петля клиента MQTT (для проверки наличия сообщений в подписанных топиках) */
  if (mqttCredentialsPresent) {
    mqtt_client.loop();
  }

  ElegantOTA.loop();

}


void wifiConnectionSetup() {
  Serial.print(F("Wifi mode is: "));
  Serial.println(WiFi.getMode());
  if (WiFi.status() == WL_CONNECTED) {
    bitWrite(errors_flags, NO_WIFI, LOW);
    if (WiFi.getMode() == WIFI_MODE_AP || WiFi.getMode() == WIFI_MODE_APSTA) {
      Serial.println(F("Disabling own AP"));
      WiFi.mode(WIFI_MODE_STA);
      WiFi.enableAP(false);
      WiFi.softAPdisconnect();
      WiFi.softAPdisconnect(false);
    }
    return;
  }
  bitWrite(errors_flags, NO_WIFI, HIGH);
  /* настраиваем беспроводные подключения */
  /* если имя сети и пароль к ней не пустые, */
  if (!ssid.isEmpty() && !password.isEmpty()) {
    /* ... включаем режим "точка доступа + клиент" */
    WiFi.mode(WIFI_AP_STA);
    /* ... и сканируем сети вокруг себя */
    Serial.println(F("Available WiFi networks scan..."));
    uint8_t n = WiFi.scanNetworks();
    /* если сетей вокруг не нашлось вообще (найдено 0 сетей), */
    if (n == 0) {
      Serial.println(F("No WiFi networks found, starting own AP."));
      /* ... активируем собственную точку доступа */
      starOwnWifiAp();
    } else {
      int nameHits = 0;
      Serial.print(F("***** "));
      Serial.print(n);
      Serial.println(F(" networks found *****"));
      for (int i = 0; i < n; ++i) {
        Serial.print(i + 1);
        Serial.print(": ");
        Serial.print(WiFi.SSID(i));
        Serial.print(F(" , signal strength: "));
        Serial.print(WiFi.RSSI(i));
        Serial.println(F(" dBi"));
        delay(10);
        if (WiFi.SSID(i) == ssid.c_str()) {
          nameHits++;
        }
      }
      if (nameHits > 0) {
        if (ip0 != 0) {
          IPAddress staticIP( ip0 , ip1 , ip2, ip3);
          IPAddress gateway( ip0 , ip1, ip2, 1);
          IPAddress subnet(255, 255, 255, 0);
          IPAddress dns(ip0, ip1, ip2, 1);
          if (WiFi.config(staticIP, gateway, subnet, dns, dns) == false) {
            Serial.println(F("Static IP configuration failed."));
          }
        } else {
          WiFi.config(INADDR_NONE, INADDR_NONE, INADDR_NONE, INADDR_NONE);
        }
        Serial.println(chipId);
        WiFi.setHostname(chipId.c_str()); //define hostname
        uint8_t tryToConnect = 0;
        Serial.print(F("Connecting to WiFi network "));
        Serial.println(ssid);
        Serial.print(F("Wait..."));
        WiFi.begin(ssid.c_str(), password.c_str());
        while (WiFi.status() != WL_CONNECTED && tryToConnect < 100 ) {
          Serial.print('.');
          delay(200);
          tryToConnect++;
        }
        if (WiFi.status() == WL_CONNECTED) {
          WiFi.setAutoReconnect(true);
          bitWrite(errors_flags, NO_WIFI, LOW);
          Serial.print(F(".\nSystem got IP address "));
          Serial.println(WiFi.localIP());
        } else {
          bitWrite(errors_flags, NO_WIFI, HIGH);
          Serial.println(F("\nCould not connect to WiFi, possible wrong password!"));
        }
      } else {
        Serial.println(F("There is no WiFi network with known credentials, starting own AP."));
        bitWrite(errors_flags, NO_WIFI, HIGH);
        starOwnWifiAp();
      }
    }
  } else {
    bitWrite(errors_flags, NO_WIFI, HIGH);
    Serial.println(F("System don`t have WiFi credentials saved, starting own AP."));
    starOwnWifiAp();
  }
}

void starOwnWifiAp() {
  if (  WiFi.getMode() == WIFI_MODE_AP ) {
    Serial.println(F("Already AP enabled"));
    return;
  }
  WiFi.mode(WIFI_MODE_AP);
  WiFi.softAP(ap_ssid.c_str(), ap_password.c_str());
}
void setLedColorWhite() {
  leds[0] = 0x040404;
  FastLED.show();
}
void setLedColorRed() {
  leds[0] = 0x000400;
  FastLED.show();
}
void setLedColorBlue() {
  leds[0] = 0x000004;
  FastLED.show();
}
void setLedColorGreen() {
  leds[0] = 0x040000;
  FastLED.show();
}

/* запись в "preferences" параметров подключения к беспроводной сети */
void saveWifiCredentials() {
  preferences.begin("all-settings", false); // true - только чтение, false - и чтение и запись
  Serial.print(F("Saving the new WiFi SSID - "));
  Serial.println(ssid);
  preferences.putString("ssid", ssid);
  Serial.print(F("Saving the new WiFi Password - "));
  Serial.println(password);
  preferences.putString("password", password);
  preferences.putUShort("ip3", ip3);
  preferences.putUShort("ip2", ip2);
  preferences.putUShort("ip1", ip1);
  preferences.putUShort("ip0", ip0);
  preferences.putString("udpAddress", udpAddress);
  Serial.println(F("System will reboot now!"));
  needSaveCredentials = false;
  preferences.end();
  delay(1000);
  ESP.restart();
}
/* запись расписания в "preferences" (ЭНЗУ) */
void saveScheduleRecords() {
  preferences.begin("all-settings", false); // true - только чтение, false - и чтение и запись
  Serial.println(F("Saving the Schedule"));
  preferences.putBytes("work-schedule", work_shedule, sizeof(work_shedule));
  preferences.end();
  needSaveSchedule = false;
}
void saveMainSettings() {
  preferences.begin("all-settings", false); // true - только чтение, false - и чтение и запись
  Serial.print(F("Saving the working mode: "));
  Serial.println(availableModes[mainMode]);
  preferences.putUChar("mainMode", mainMode);
  Serial.print(F("Saving hysteresis: "));
  Serial.println(hysteresis);
  preferences.putFloat("hysteresis", hysteresis);
  Serial.print(F("Saving Force-On temperature: "));
  Serial.println(forcetemp);
  preferences.putFloat("forcetemp", forcetemp);
  Serial.print(F("Saving Alarm Overheat temperature: "));
  Serial.println(alarmtemp);
  preferences.putFloat("alarmtemp", alarmtemp);
  preferences.putString("devicename", devicename);
  preferences.end();
  needSaveSettings = false;
}
void saveMqttCredentials() {
  preferences.begin("all-settings", false); // true - только чтение, false - и чтение и запись
  Serial.print(F("Saving MQTT-server address: "));
  Serial.println(mqttserver);
  preferences.putString("mqttserver", mqttserver);
  Serial.print(F("Saving MQTT-server user name: "));
  Serial.println(mqttuser);
  preferences.putString("mqttuser", mqttuser);
  Serial.print(F("Saving MQTT-server password: "));
  Serial.println(mqttpassword);
  preferences.putString("mqttpassword", mqttpassword);
  preferences.putString("logger", logger);
  preferences.end();
  if (!mqttserver.isEmpty() && !mqttuser.isEmpty() && !mqttpassword.isEmpty()) {
    mqttCredentialsPresent = true;
  } else {
    mqttCredentialsPresent = false;
  }
  needSaveMqttCredentials = false;
}
void savePowerConsumption() {
  preferences.begin("all-settings", false); // true - только чтение, false - и чтение и запись
  preferences.putULong64("powercntr", powerOnTimeTotal);
  Serial.print(F("Saving total workin time: "));
  Serial.println(formattedTimeFromInt(powerOnTimeTotal));
  preferences.end();
  needSaveTotalTime = false;
}
/*
   Раз в 10 секунд проверяем - подключен ли вайфай. Если да, вызываем проверку
   того, работает ли NTP-клиент. После этого - проверяем текущий режим работы
   вайфай (если 3, то это Station + softAP, если 1 - то Station). Если 3-й, то
   выключаем собственную точку доступа, и плата работает в режиме "чистого"
   клиента, подключееногок внешней точке доступа.
*/
void tenSecondProcedure() {
  /* watchdog timer reset procedure (must be run often than WDT period) */
  esp_task_wdt_reset();
  /* проверка подключения к WiFi и NTP-синхронизации */
  if (WiFi.status() == WL_CONNECTED) {
    ntpTimeCheckAndSet();
    if (WiFi.getMode() == 3) {
      WiFi.mode(WIFI_MODE_STA);
      WiFi.enableAP(false);
      WiFi.softAPdisconnect();
      WiFi.softAPdisconnect(false);
    }
  } else {
    bitWrite(errors_flags, NO_WIFI, HIGH);
    wifiConnectionSetup();
  }

  if (powerState) {
    powerOnTimeTotal = powerOnTimeTotal + 10;
    if (powerOnTimeTotal - powerOnTimePrevious > 3600) {
      powerOnTimePrevious = powerOnTimeTotal;
      Serial.println(F("PowerOn-time differ too much, need to save it to EEPROM!"));
      needSaveTotalTime = true;
    }
  }

  /* проверка - не нажата ли кнопка принудительного включения оборудования */
  if (digitalRead(force_button_pin) == LOW) {
    if ((mainMode == 2) or
        (bitRead(errors_flags, NO_TEMP_SENS_W) == 0
         and bitRead(errors_flags, NO_TEMP_SENS_A) == 0
         and bitRead(errors_flags, BOILER_OVERHEAT) == 0)) {
      forceOnTimerCounter = FORCE_ON_TIMER_STEP;
      Serial.println(F("Force-On button pressed!"));
      messages_quene.push(4);
    }
  }

  if (mainMode != 2) {
    currentTempWorking = getMainTemperature();
    if (currentTempWorking == DEVICE_DISCONNECTED_C) {
      if (bitRead(errors_flags, NO_TEMP_SENS_W) == 0) {
        messages_quene.push(0);
        /* "АВАРИЯ - Нет сигнала от основного датчика температуры! Оборудование выключено"; */
      }
      bitWrite(errors_flags, NO_TEMP_SENS_W, HIGH);
      errorMessageText = "Не исправен основной датчик температуры!";
      if (powerState) {
        powerState = false;
        needSaveTotalTime = true;
      }
    } else {
      bitWrite(errors_flags, NO_TEMP_SENS_W, LOW);
    }
    currentTempAlarm = getAlarmTemperature();
    if (currentTempAlarm == DEVICE_DISCONNECTED_C) {
      if (bitRead(errors_flags, NO_TEMP_SENS_A) == 0) {
        messages_quene.push(1);
        /* "АВАРИЯ - Нет сигнала от аварийного датчика температуры! Оборудование выключено!"; */
      }
      bitWrite(errors_flags, NO_TEMP_SENS_A, HIGH);
      errorMessageText = "Не исправен аварийный датчик температуры!";
      if (powerState) {
        powerState = false;
        needSaveTotalTime = true;
      }
    } else {
      bitWrite(errors_flags, NO_TEMP_SENS_A, LOW);
    }
    if (currentTempAlarm >= alarmtemp || currentTempWorking >= alarmtemp) {
      if (bitRead(errors_flags, BOILER_OVERHEAT) == 0) {
        messages_quene.push(2);
        /*"АВАРИЯ - Температура превысила аврийный порог! Оборудование выключено!";*/
      }
      bitWrite(errors_flags, BOILER_OVERHEAT, HIGH);
      errorMessageText = "АВАРИЯ - перегрев воды в баке!";
      if (powerState) {
        powerState = false;
        needSaveTotalTime = true;
      }
    } else if (bitRead(errors_flags, BOILER_OVERHEAT) == 1 && currentTempAlarm < ( alarmtemp / 2 ) ) {
      bitWrite(errors_flags, BOILER_OVERHEAT, LOW);
    }
  } else {
    bitWrite(errors_flags, NO_TEMP_SENS_W, LOW);
    bitWrite(errors_flags, NO_TEMP_SENS_A, LOW);
    bitWrite(errors_flags, BOILER_OVERHEAT, LOW);
  }

  checkScheduleHits();

  checkTemperature();

  if (powerState) {
    setLedColorBlue();
  } else if (errors_flags != 0) {
    setLedColorRed();
    String a;
    for (int i = 7; i >= 0; i--) {
      if (bitRead(errors_flags, i) == 1) {
        a += "1";
      } else {
        a += "0";
      }
    }
    Serial.print(F("Sysmem not ready! Error code: "));
    Serial.println( a );
  } else {
    setLedColorGreen();
  }

  one_min_counter = ( one_min_counter + 1 ) % 6;      // подсчитываем 1 минуту
  if (one_min_counter == 0) {                         // если прошла одна минута...
    if (bitRead(errors_flags, NO_WIFI) == 0 ) {       // ...и мы подключены к вайфай,
      mqtt_powerstate_send();                           // сбрасываем данные по
      mqtt_temperature_send();                          // протоколу mqtt
    }
  }
  ten_min_counter = ( ten_min_counter + 1 ) % 60 ;    // подсчитываем 10 минут
  if (ten_min_counter == 0) {                         // если прошло 10 минут...
    if (bitRead(errors_flags, NO_WIFI) == 0 ) {       // ...и мы подключены к вайфай,
      sendDataToCollector();                          // сбрасываем данные в сервер-коллектор
    }
  }

  if (forceOnTimerCounter > 0) {
    forceOnTimerCounter = forceOnTimerCounter - 10;
  }

  bool cur_state_of_relay_pin = digitalRead(relay_control_pin);
  if (powerState && !cur_state_of_relay_pin) {
    digitalWrite(relay_control_pin, HIGH);
  } else if (!powerState && cur_state_of_relay_pin) {
    digitalWrite(relay_control_pin, LOW);
  }

  /* сбрасываем "флаг" 10 секунд */
  portENTER_CRITICAL(&timer0Mux);
  tenSecondFlag = false;
  portEXIT_CRITICAL(&timer0Mux);
}

float getMainTemperature() {
  float temporaryResult;
  for (int i = 0; i < 10; i++) {
    delay(3);
    main_sensor.requestTemperatures();
    temporaryResult = main_sensor.getTempCByIndex(0);
    if (temporaryResult != DEVICE_DISCONNECTED_C) {
      break;
    }
  }
  return temporaryResult;
}

float getAlarmTemperature() {
  float temporaryResult;
  for (int i = 0; i < 10; i++) {
    delay(3);
    alarm_sensor.requestTemperatures();
    temporaryResult = alarm_sensor.getTempCByIndex(0);
    if (temporaryResult != DEVICE_DISCONNECTED_C) {
      break;
    }
  }
  return temporaryResult;
}

void sendDataToCollector() {
  if (WiFi.status() != WL_CONNECTED  or logger.isEmpty() or currentTempWorking == DEVICE_DISCONNECTED_C) {
    return;
  }
  String fullPath = "http://" + logger + collectorSiteAddress + "devid=" + chipId + "&temp=" + currentTempWorking + "&humid=" + powerState;
  http.begin(fullPath.c_str());
  int httpResponseCode = http.GET();
  // Free resources*/
  http.end();
}

String formattedTimeFromInt(uint64_t elaspedTime) {
  String answer = "";
  int h = elaspedTime / 60 / 60;
  int m = (elaspedTime / 60 ) % 60;
  int s = elaspedTime % 60;
  answer += String(h);
  answer += " час. ";
  if (m < 10) {
    answer += "0";
  }
  answer += String(m);
  answer += " мин. ";
  if (s < 10) {
    answer += "0";
  }
  answer += String(s);
  answer += " сек.";
  return answer;
}

void checkTemperature() {
  if (!powerControlAllowed) {
    return;
  }
  float temperature_from_schedule;
  if (mainMode == 2) {
    temperature_from_schedule = 129.00 ;
  } else if (forceOnTimerCounter > 0) {
    temperature_from_schedule = forcetemp ;
  } else {
    temperature_from_schedule = work_shedule[selectedScheduleRow].desired_temp;
  }
  Serial.print(F("Comparing current temperature ("));
  Serial.print(currentTempWorking);
  Serial.print(F(") and temperature from schedule ("));
  Serial.print(temperature_from_schedule);
  Serial.print(F("). "));
  Serial.printf("Current working mode is: \"%s\".\n", availableModes[mainMode]);
  if ((mainMode == 0 and currentTempWorking > temperature_from_schedule)
      or (mainMode == 1 and currentTempWorking < temperature_from_schedule)) {
    if (powerState) {
      Serial.println(F("Пора выключить оборудование"));
      powerState = false;
      needSaveTotalTime = true;
      messages_quene.push(8);
    } else {
      Serial.println(F("Оборудование уже выключено"));
    }
  } else if ((mainMode == 0 and currentTempWorking < (temperature_from_schedule - hysteresis))
             or (mainMode == 1 and currentTempWorking > (temperature_from_schedule + hysteresis))
             or (mainMode == 2 and currentTempWorking < 100.00) ) {
    if (!powerState) {
      Serial.println(F("Необходимо включить оборудование"));
      powerState = true;
      messages_quene.push(9);
    } else {
      Serial.println(F("Оборудование уже включено"));
    }
  }
}

void checkScheduleHits() {
  if ((forceOnTimerCounter > 0 and mainMode == 2) or
      (forceOnTimerCounter > 0
       and bitRead(errors_flags, NO_TEMP_SENS_W) == 0
       and bitRead(errors_flags, NO_TEMP_SENS_A) == 0
       and bitRead(errors_flags, BOILER_OVERHEAT) == 0)) {
    powerControlAllowed = true;
    return;
  }
  if ( errors_flags > 0 ) {
    powerState = false;
    powerControlAllowed = false;
    selectedScheduleRow = -1;
    return;
  }

  uint16_t curTimeForCompare = curDayTimeToInt();
  Serial.print(F("Schedule check at "));
  Serial.print(intCurDayTimeToString(curTimeForCompare));
  Serial.printf(" (%d).\n", curTimeForCompare);
  /* когда ни одно из заданий расписания не выполняется, проверяем массив шедулера, не пора ли включать какое-то задание */
  if (selectedScheduleRow < 0) {
    for (int8_t i = 0 ; i < sched_rows_numb ; i++ ) {
      if (work_shedule[i].isactiv != true ) {
        continue;
      }
      uint16_t start_time_from_schedule = work_shedule[i].start_time;
      uint16_t stop_time_from_schedule = work_shedule[i].stop_time;
      if (start_time_from_schedule != stop_time_from_schedule) {
        if (stop_time_from_schedule < start_time_from_schedule) {
          stop_time_from_schedule = stop_time_from_schedule + 10080 ;
          if (DateTime.getParts().getWeekDay() == 0) {
            curTimeForCompare = curTimeForCompare + 10080;
          }
        }
        Serial.printf("Comparing current-time %d and schedule-ON-time %d \n", curTimeForCompare, start_time_from_schedule);
        if ( curTimeForCompare >= start_time_from_schedule &&  curTimeForCompare < stop_time_from_schedule ) {
          selectedScheduleRow = i;
          Serial.printf("Наступило время для выполнения задачи %d \n", selectedScheduleRow);
          powerControlAllowed = true;
          messages_quene.push(6);
          /* "Согласно расписания ВКЛЮЧЕН режим поддержания заданной температуры"; */
          return;
        }
      }
    }
    Serial.println(F("Ни одно из заданий расписания не подошло."));
    /* если же какое-то из заданий расписания уже в работе, проверяем, не пора ли его выключать */
  } else {
    Serial.printf("Выполняется задача %d \n", selectedScheduleRow);
    uint16_t start_time_from_schedule = work_shedule[selectedScheduleRow].start_time;
    uint16_t stop_time_from_schedule = work_shedule[selectedScheduleRow].stop_time;
    Serial.printf("Comparing current-time %d and schedule-OFF-time %d \n", curTimeForCompare, stop_time_from_schedule);
    if ((work_shedule[selectedScheduleRow].isactiv == false)
        or (stop_time_from_schedule > start_time_from_schedule
            && curTimeForCompare >= stop_time_from_schedule )
        or (stop_time_from_schedule < start_time_from_schedule
            && curTimeForCompare >= stop_time_from_schedule
            && curTimeForCompare < start_time_from_schedule)) {
      Serial.printf("Время задачи %d истекло, выключаем ее.\n", selectedScheduleRow);
      messages_quene.push(7);
      /*"Согласно расписания режим поддержания заданной температуры ВЫКЛЮЧЕН"; */
      selectedScheduleRow = -1;
      powerState = false;
      powerControlAllowed = false;
    }
  }
}

/* преобразование в "штамп" ТЕКУЩЕГО дня недели и времени */
int curDayTimeToInt() {
  DateTimeParts p = DateTime.getParts();
  uint16_t curVal = p.getWeekDay() * 1440 + p.getHours() * 60 + p.getMinutes();
  return curVal;
}
/* преобразование в "штамп" ПРОИЗВОЛЬНОГО дня недели и времени */
int anyDayTimeToInt(uint8_t day, uint8_t hour, uint8_t minute) {
  uint16_t retVal = (day * 1440) + (hour * 60) + minute;
  return retVal;
}

/* преобразование "штапма" в строку с указанением дня недели и времени */
String intCurDayTimeToString(uint16_t intValue) {
  String answer;
  uint8_t day = intValue / 24 / 60 ;
  uint8_t hour = (intValue / 60) % 24;
  uint8_t minute = intValue % 60;
  answer += daysOfTheWeek[day];
  answer += " ";
  if (hour < 10) {
    answer += "0";
  }
  answer += hour;
  answer += ":";
  if (minute < 10) {
    answer += "0";
  }
  answer += minute;
  return answer;
}

/*
   Проверка того, связался ли NTP-клиент с удаленным сервером и получил ли от
   него значение текущего времени. Если нет, вызываем процедуру настройки и
   запуска NTP-клиента. Если да - проверяем флаг "система только что загрузилась",
   и если он установлен, отсылаем соответствующее лог-сообщение
*/
void ntpTimeCheckAndSet() {
  if (!DateTime.isTimeValid()) {
    bitWrite(errors_flags, NO_NTP, HIGH);
    Serial.println(F("Time from NTP-server is not yet available..."));
    setupDateTime();
  } else {
    bitWrite(errors_flags, NO_NTP, LOW);
    if (bootAfterPowerLost && DateTime.now() > 900000000) {
      Serial.printf("Local Time: %s\n", DateTime.toString().c_str());
      Serial.println(F("Trying to send \"first after boot\" SYSLOG-message"));
      messages_quene.unshift(3);
      bootAfterPowerLost = false;
    }
  }
}

/* setup and start NTP time reading (after wifi connected) */
void setupDateTime() {
  DateTime.setServer("pool.ntp.org");
  DateTime.setTimeZone("EET-2EEST,M3.5.0/3,M10.5.0/4");
  DateTime.begin();
}

/* обработчик "сообщений-команд", поступающих от сервера MQTT */
void receivedCallback(char* topic, byte* payload, unsigned int length) {
  Serial.print(F("Message received: "));
  Serial.println(topic);
  Serial.print(F("payload: "));
  for (int i = 0; i < length; i++) {
    Serial.print((char)payload[i]);
  }
  Serial.println();
  if (strcmp(topic, MQTT_PUB_FORCE_ON) == 0) {  // проверяем топик
    if ((char)payload[0] == '1') {              // проверяем сообщение
      if (errors_flags == 0) {
        forceOnTimerCounter = forceOnTimerCounter + FORCE_ON_TIMER_STEP;
        messages_quene.push(4);
      } else {
        messages_quene.push(5);
      }
    }
  }
  /*  if (strcmp(topic, topic_mode_set) == 0) {
      if ((char)payload[0] == '0') {

      } else if ((char)payload[0] == '1') {

      } else if ((char)payload[0] == '2') {

      }
    } */
}

/* отправка в сервер MQTT лог-сообщений */
uint8_t mqtt_log_message_send() {
  String wholeMessage = DateTime.toString() + " :: " + mqttMessageToSend;
  mqtt_server_connect();
  uint8_t packetIdPub1 = mqtt_client.publish(MQTT_PUB_LOG, wholeMessage.c_str(), true);
  wholeMessage = String();
  return packetIdPub1;
}

/* отправка в сервер MQTT текущей температуры */
void mqtt_temperature_send() {
  if (bitRead(errors_flags, NO_TEMP_SENS_W) == 1) {
    return;
  }
  char msg_out[20];
  dtostrf(currentTempWorking, 2, 2, msg_out);
  mqtt_server_connect();
  if (mqttCredentialsPresent) {
    uint16_t packetIdPub1 = mqtt_client.publish(MQTT_PUB_TEMP, msg_out);
  }
}

/* отправка в сервер MQTT текущего состояния питания оборудования */
void mqtt_powerstate_send() {
  mqtt_server_connect();
  if (mqttCredentialsPresent) {
    uint16_t packetIdPub1 = mqtt_client.publish(MQTT_PUB_POWERSTATE, ((powerState) ? "1" : "0"));
  }
}

/* подключение к серверу MQTT */
void mqtt_server_connect() {
  if (mqttCredentialsPresent) {
    if (!mqtt_client.connected()) {
      Serial.print(F("MQTT server not yet connected now, rc="));
      Serial.println(mqtt_client.state());
      Serial.println(F("Trying to connect MQTT server now..."));
      if (mqtt_client.connect(chipId.c_str(), mqttuser.c_str(), mqttpassword.c_str())) {
        Serial.println(F("MQTT server connected!"));
        mqtt_client.subscribe(MQTT_PUB_LOG);
        mqtt_client.subscribe(MQTT_PUB_TEMP);
        mqtt_client.subscribe(MQTT_PUB_POWERSTATE);
        mqtt_client.subscribe(MQTT_PUB_FORCE_ON);
      } else {
        Serial.print(F("Connection failed, rc="));
        Serial.print(mqtt_client.state());
        Serial.println(F(" try again later..."));
      }
    }
  }
}

/* процедура отправки лог-сообщения на внешний сервер. Если в настройках
   есть адрес сервера (udpAddress), то сообщение отсылается на его порт
   (udpPort). Если адреса нет, то отправка не осуществляется. В конце
   процедуры само сообщение сбрасывается (в "пустую строку").
*/
void syslog_message_send() {
  udp.beginPacket(udpAddress.c_str(), udpPort);
  udp.print(chipId + ": " + logMessageToSend);
  udp.endPacket();
  logMessageToSend = String();
}
